import logging

import six
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from dagster_ssh.resources import SSHResource, key_from_str

from dagster.seven import mock


def generate_ssh_key():
    # generate private/public key pair
    key = rsa.generate_private_key(backend=default_backend(), public_exponent=65537, key_size=2048)

    # get private key in PEM container format
    return six.ensure_str(
        key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
    )


@mock.patch('paramiko.SSHClient')
def test_ssh_connection_with_password(ssh_mock):
    ssh_resource = SSHResource(
        remote_host='remote_host',
        remote_port=12345,
        username='username',
        password='password',
        key_file='fake.file',
        timeout=10,
        keepalive_interval=30,
        compress=True,
        no_host_key_check=False,
        allow_host_key_change=False,
        logger=logging.root.getChild('test_resources'),
    )

    with ssh_resource.get_connection():
        ssh_mock.return_value.connect.assert_called_once_with(
            hostname='remote_host',
            username='username',
            password='password',
            pkey=None,
            key_filename='fake.file',
            timeout=10,
            compress=True,
            port=12345,
            sock=None,
            look_for_keys=False,
        )


@mock.patch('paramiko.SSHClient')
def test_ssh_connection_without_password(ssh_mock):
    ssh_resource = SSHResource(
        remote_host='remote_host',
        remote_port=12345,
        username='username',
        password=None,
        timeout=10,
        key_file='fake.file',
        keepalive_interval=30,
        compress=True,
        no_host_key_check=False,
        allow_host_key_change=False,
        logger=logging.root.getChild('test_resources'),
    )

    with ssh_resource.get_connection():
        ssh_mock.return_value.connect.assert_called_once_with(
            hostname='remote_host',
            username='username',
            pkey=None,
            key_filename='fake.file',
            timeout=10,
            compress=True,
            port=12345,
            sock=None,
        )


@mock.patch('paramiko.SSHClient')
def test_ssh_connection_with_key_string(ssh_mock):
    ssh_key = generate_ssh_key()

    ssh_resource = SSHResource(
        remote_host='remote_host',
        remote_port=12345,
        username='username',
        password=None,
        timeout=10,
        key_string=six.ensure_str(ssh_key),
        keepalive_interval=30,
        compress=True,
        no_host_key_check=False,
        allow_host_key_change=False,
        logger=logging.root.getChild('test_resources'),
    )

    with ssh_resource.get_connection():
        ssh_mock.return_value.connect.assert_called_once_with(
            hostname='remote_host',
            username='username',
            key_filename=None,
            pkey=key_from_str(ssh_key),
            timeout=10,
            compress=True,
            port=12345,
            sock=None,
        )


@mock.patch('dagster_ssh.resources.SSHTunnelForwarder')
def test_tunnel_with_password(ssh_mock):
    ssh_resource = SSHResource(
        remote_host='remote_host',
        remote_port=12345,
        username='username',
        password='password',
        timeout=10,
        key_file='fake.file',
        keepalive_interval=30,
        compress=True,
        no_host_key_check=False,
        allow_host_key_change=False,
        logger=logging.root.getChild('test_resources'),
    )

    with ssh_resource.get_tunnel(1234):
        ssh_mock.assert_called_once_with(
            'remote_host',
            ssh_port=12345,
            ssh_username='username',
            ssh_password='password',
            ssh_pkey='fake.file',
            ssh_proxy=None,
            local_bind_address=('localhost',),
            remote_bind_address=('localhost', 1234),
            logger=ssh_resource.log,
        )


@mock.patch('dagster_ssh.resources.SSHTunnelForwarder')
def test_tunnel_without_password(ssh_mock):
    ssh_resource = SSHResource(
        remote_host='remote_host',
        remote_port=12345,
        username='username',
        password=None,
        timeout=10,
        key_file='fake.file',
        keepalive_interval=30,
        compress=True,
        no_host_key_check=False,
        allow_host_key_change=False,
        logger=logging.root.getChild('test_resources'),
    )

    with ssh_resource.get_tunnel(1234):
        ssh_mock.assert_called_once_with(
            'remote_host',
            ssh_port=12345,
            ssh_username='username',
            ssh_pkey='fake.file',
            ssh_proxy=None,
            local_bind_address=('localhost',),
            remote_bind_address=('localhost', 1234),
            host_pkey_directories=[],
            logger=ssh_resource.log,
        )


@mock.patch('dagster_ssh.resources.SSHTunnelForwarder')
def test_tunnel_with_string_key(ssh_mock):
    ssh_key = generate_ssh_key()

    ssh_resource = SSHResource(
        remote_host='remote_host',
        remote_port=12345,
        username='username',
        password=None,
        timeout=10,
        key_string=six.ensure_str(ssh_key),
        keepalive_interval=30,
        compress=True,
        no_host_key_check=False,
        allow_host_key_change=False,
        logger=logging.root.getChild('test_resources'),
    )

    with ssh_resource.get_tunnel(1234):
        ssh_mock.assert_called_once_with(
            'remote_host',
            ssh_port=12345,
            ssh_username='username',
            ssh_pkey=key_from_str(ssh_key),
            ssh_proxy=None,
            local_bind_address=('localhost',),
            remote_bind_address=('localhost', 1234),
            host_pkey_directories=[],
            logger=ssh_resource.log,
        )



import getpass
import os

import paramiko
from paramiko.config import SSH_PORT
from six import StringIO
from sshtunnel import SSHTunnelForwarder

from dagster import Field, check, resource
from dagster.utils import merge_dicts


def key_from_str(key_str):
    '''Creates a paramiko SSH key from a string.'''
    check.str_param(key_str, 'key_str')

    # py2 StringIO doesn't support with
    key_file = StringIO(key_str)
    result = paramiko.RSAKey.from_private_key(key_file)
    key_file.close()
    return result


class SSHResource(object):
    '''
    Resource for ssh remote execution using Paramiko.
    ref: https://github.com/paramiko/paramiko
    '''

    def __init__(
        self,
        remote_host,
        remote_port,
        username=None,
        password=None,
        key_file=None,
        key_string=None,
        timeout=10,
        keepalive_interval=30,
        compress=True,
        no_host_key_check=True,
        allow_host_key_change=False,
        logger=None,
    ):
        self.remote_host = check.str_param(remote_host, 'remote_host')
        self.remote_port = check.opt_int_param(remote_port, 'remote_port')
        self.username = check.opt_str_param(username, 'username')
        self.password = check.opt_str_param(password, 'password')
        self.key_file = check.opt_str_param(key_file, 'key_file')
        self.timeout = check.opt_int_param(timeout, 'timeout')
        self.keepalive_interval = check.opt_int_param(keepalive_interval, 'keepalive_interval')
        self.compress = check.opt_bool_param(compress, 'compress')
        self.no_host_key_check = check.opt_bool_param(no_host_key_check, 'no_host_key_check')
        self.allow_host_key_change = check.opt_bool_param(
            allow_host_key_change, 'allow_host_key_change'
        )
        self.log = logger

        self.host_proxy = None

        # Create RSAKey object from private key string
        self.key_obj = key_from_str(key_string) if key_string is not None else None

        # Auto detecting username values from system
        if not self.username:
            logger.debug(
                'username to ssh to host: %s is not specified. Using system\'s default provided by'
                ' getpass.getuser()' % self.remote_host
            )
            self.username = getpass.getuser()

        user_ssh_config_filename = os.path.expanduser('~/.ssh/config')
        if os.path.isfile(user_ssh_config_filename):
            ssh_conf = paramiko.SSHConfig()
            ssh_conf.parse(open(user_ssh_config_filename))
            host_info = ssh_conf.lookup(self.remote_host)
            if host_info and host_info.get('proxycommand'):
                self.host_proxy = paramiko.ProxyCommand(host_info.get('proxycommand'))

            if not (self.password or self.key_file):
                if host_info and host_info.get('identityfile'):
                    self.key_file = host_info.get('identityfile')[0]

    def get_connection(self):
        """
        Opens a SSH connection to the remote host.

        :rtype: paramiko.client.SSHClient
        """
        client = paramiko.SSHClient()
        if not self.allow_host_key_change:
            self.log.warning(
                'Remote Identification Change is not verified. This won\'t protect against '
                'Man-In-The-Middle attacks'
            )
            client.load_system_host_keys()
        if self.no_host_key_check:
            self.log.warning(
                'No Host Key Verification. This won\'t protect against Man-In-The-Middle attacks'
            )
            # Default is RejectPolicy
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        if self.password and self.password.strip():
            client.connect(
                hostname=self.remote_host,
                username=self.username,
                password=self.password,
                key_filename=self.key_file,
                pkey=self.key_obj,
                timeout=self.timeout,
                compress=self.compress,
                port=self.remote_port,
                sock=self.host_proxy,
                look_for_keys=False,
            )
        else:
            client.connect(
                hostname=self.remote_host,
                username=self.username,
                key_filename=self.key_file,
                pkey=self.key_obj,
                timeout=self.timeout,
                compress=self.compress,
                port=self.remote_port,
                sock=self.host_proxy,
            )

        if self.keepalive_interval:
            client.get_transport().set_keepalive(self.keepalive_interval)

        return client

    def get_tunnel(self, remote_port, remote_host='localhost', local_port=None):
        check.int_param(remote_port, 'remote_port')
        check.str_param(remote_host, 'remote_host')
        check.opt_int_param(local_port, 'local_port')

        if local_port is not None:
            local_bind_address = ('localhost', local_port)
        else:
            local_bind_address = ('localhost',)

        # Will prefer key string if specified, otherwise use the key file
        pkey = self.key_obj if self.key_obj else self.key_file

        if self.password and self.password.strip():
            client = SSHTunnelForwarder(
                self.remote_host,
                ssh_port=self.remote_port,
                ssh_username=self.username,
                ssh_password=self.password,
                ssh_pkey=pkey,
                ssh_proxy=self.host_proxy,
                local_bind_address=local_bind_address,
                remote_bind_address=(remote_host, remote_port),
                logger=self.log,
            )
        else:
            client = SSHTunnelForwarder(
                self.remote_host,
                ssh_port=self.remote_port,
                ssh_username=self.username,
                ssh_pkey=pkey,
                ssh_proxy=self.host_proxy,
                local_bind_address=local_bind_address,
                remote_bind_address=(remote_host, remote_port),
                host_pkey_directories=[],
                logger=self.log,
            )

        return client


@resource(
    {
        'remote_host': Field(str, description='remote host to connect', is_required=True),
        'remote_port': Field(
            int,
            description='port of remote host to connect (Default is paramiko SSH_PORT)',
            is_required=False,
            default_value=SSH_PORT,
        ),
        'username': Field(
            str, description='username to connect to the remote_host', is_required=False
        ),
        'password': Field(
            str,
            description='password of the username to connect to the remote_host',
            is_required=False,
        ),
        'key_file': Field(
            str, description='key file to use to connect to the remote_host.', is_required=False
        ),
        'key_string': Field(
            str, description='key string to use to connect to remote_host', is_required=False
        ),
        'timeout': Field(
            int,
            description='timeout for the attempt to connect to the remote_host.',
            is_required=False,
            default_value=10,
        ),
        'keepalive_interval': Field(
            int,
            description='send a keepalive packet to remote host every keepalive_interval seconds',
            is_required=False,
            default_value=30,
        ),
        'compress': Field(bool, is_required=False, default_value=True),
        'no_host_key_check': Field(bool, is_required=False, default_value=True),
        'allow_host_key_change': Field(bool, is_required=False, default_value=False),
    }
)
def ssh_resource(init_context):
    args = init_context.resource_config
    args = merge_dicts(init_context.resource_config, {'logger': init_context.log_manager})
    return SSHResource(**args)



import os

from dagster import Enum, EnumValue, Field, solid
from dagster.utils import mkdir_p


@solid(
    config={
        'local_filepath': Field(
            str, is_required=True, description='local file path to get or put.'
        ),
        'remote_filepath': Field(
            str, is_required=True, description='remote file path to get or put.'
        ),
        'operation': Field(
            Enum('SFTPOperation', [EnumValue('GET'), EnumValue('PUT')]),
            is_required=False,
            default_value='PUT',
            description='specify operation \'GET\' or \'PUT\', defaults to PUT',
        ),
        'confirm': Field(
            bool,
            is_required=False,
            default_value=True,
            description='specify if the SFTP operation should be confirmed, defaults to True',
        ),
    },
    required_resource_keys={'ssh_resource'},
)
def sftp_solid(context):
    '''
    Ported from Airflow's SFTPOperator.

    sftp_solid: for transferring files from remote host to local or vice a versa. This solid uses
    ssh_resource to open a SFTP transport channel that serve as basis for file transfer.
    '''
    local_filepath = context.solid_config.get('local_filepath')
    remote_filepath = context.solid_config.get('remote_filepath')
    operation = context.solid_config.get('operation')
    confirm = context.solid_config.get('confirm')

    with context.resources.ssh_resource.get_connection() as ssh_client:
        sftp_client = ssh_client.open_sftp()
        if operation == 'GET':
            local_folder = os.path.dirname(local_filepath)

            # Create intermediate directories if they don't exist
            mkdir_p(local_folder)

            context.log.info(
                'Starting to transfer from {0} to {1}'.format(remote_filepath, local_filepath)
            )
            sftp_client.get(remote_filepath, local_filepath)

        else:
            context.log.info(
                'Starting to transfer file from {0} to {1}'.format(local_filepath, remote_filepath)
            )

            sftp_client.put(local_filepath, remote_filepath, confirm=confirm)

    return local_filepath



from dagster.core.utils import check_dagster_package_version

from .resources import ssh_resource
from .solids import sftp_solid
from .version import __version__

check_dagster_package_version('dagster-ssh', __version__)

__all__ = ['ssh_resource', 'sftp_solid']



import os

from dagster_ssh import sftp_solid, ssh_resource

from dagster import ModeDefinition, execute_solid
from dagster.seven import get_system_temp_directory


def test_sftp_solid(sftpserver):
    tmp_path = get_system_temp_directory()
    readme_file = os.path.join(tmp_path, 'readme.txt')

    with sftpserver.serve_content({'a_dir': {'readme.txt': 'hello, world'}}):
        result = execute_solid(
            sftp_solid,
            ModeDefinition(resource_defs={'ssh_resource': ssh_resource}),
            environment_dict={
                'solids': {
                    'sftp_solid': {
                        'config': {
                            'local_filepath': readme_file,
                            'remote_filepath': 'a_dir/readme.txt',
                            'operation': 'GET',
                        }
                    }
                },
                'resources': {
                    'ssh_resource': {
                        'config': {
                            'remote_host': sftpserver.host,
                            'remote_port': sftpserver.port,
                            'username': 'user',
                            'password': 'pw',
                            'no_host_key_check': True,
                        }
                    }
                },
            },
        )
        assert result.success

    with open(readme_file, 'rb') as f:
        contents = f.read()
        assert b'hello, world' in contents
